/* 
 *	The HTE Editor
 *	htfile.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htdebug.h"
#include "htfile.h"
#include "httools.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/*
 *	CLASS htfile
 */

htfile::htfile()
{
	error=0;
}

htfile::~htfile()
{
}

ABSTRACT int htfile::read(void *buf, int size)
{
	abstract();
	return 0;
}

ABSTRACT int htfile::write(void *buf, int size)
{
	abstract();
	return 0;
}

ABSTRACT int htfile::seek(int pos)
{
	abstract();
	return 0;
}

ABSTRACT int htfile::tell()
{
	abstract();
	return 0;
}

ABSTRACT int htfile::size()
{
	abstract();
	return 0;
}

ABSTRACT char *htfile::getname()
{
	abstract();
	return 0;
}

void htfile::seterror(int e)
{
	error=e;
}

int htfile::geterror()
{
	return error;
}

STUB void htfile::flush()
{
}

STUB void htfile::invd()
{
}

STUB int htfile::dirty()
{
	return 0;
}

STUB int htfile::bytedirty(int ofs)
{
	return 0;
}

/*
 *	CLASS htrawfile
 */

htrawfile::htrawfile()
{
	name=0;
	mode=0;
	file=0;
	isdirty=0;
}

htrawfile::htrawfile(const char *name, const char *mode) : htfile()
{
	this->name=(char*)malloc(strlen(name)+1);
	strcpy(this->name, name);
	this->mode=(char*)malloc(strlen(mode)+1);
	strcpy(this->mode, mode);
	file=fopen(name, mode);
	if (!file) seterror(1);
	isdirty=0;
}

htrawfile::~htrawfile()
{
	if (name) delete name;
	if (mode) delete mode;
	if (file) fclose(file);
}

int htrawfile::read(void *buf, int size)
{
	return fread(buf, 1, size, file);
}

int htrawfile::write(void *buf, int size)
{
	isdirty=1;
	return fwrite(buf, 1, size, file);
}

int htrawfile::seek(int pos)
{
	return fseek(file, pos, SEEK_SET);
}

int htrawfile::tell()
{
	return ftell(file);
}

int htrawfile::size()
{
/* FIXME: hack to comply with ANSI rules */
	int t=tell();
	fseek(file, 0, SEEK_END);	/* zero is allowed ;-) */
	int r=tell();
	seek(t);
	return r;
}

char *htrawfile::getname()
{
	return name;
}

int htrawfile::dirty()
{
	int d=isdirty;
	isdirty=0;
	return d;
}

/*
 *	CLASS htrwfile
 */

htrwfile::htrwfile()
{
	vmod=0;
}

htrwfile::htrwfile(const char *name) : htrawfile(name, "rb")
{
	init_buf();
	frag_count=0;
}

htrwfile::~htrwfile()
{
	done_buf();
}

int htrwfile::seek(int pos)
{
	return fseek(file, pos, SEEK_SET);
}

int htrwfile::tell()
{
	return ftell(file);
}

void htrwfile::init_buf()
{
	vmod=new ht_tree();
}

void htrwfile::done_buf()
{
	if (vmod) {
		delete vmod;
		vmod=0;
	}
}

void htrwfile::flush()
{
	int prevofs=tell();
	assert(freopen(name, "ab", file));
	ht_node *d=vmod->enum_first();
	while (d) {
		seek(d->key);
		for (int i=((ht_data_page*)d->data)->start; i<=((ht_data_page*)d->data)->end; i++) {
			fputc(((ht_data_page*)d->data)->get(i), file);
		}
		d=vmod->enum_next();
	}
	invd();
	assert(freopen(name, mode, file));
	seek(prevofs);
}

void htrwfile::invd()
{
	done_buf();
	init_buf();
}

void htrwfile::putc(unsigned char c)
{
	ht_data *d;
	int ofs=tell();
	int tofs=ofs & ~(HT_PAGE_SIZE-1);
	if ((d=vmod->get(tofs))) {
		((ht_data_page*)d)->set(ofs-tofs, c);
		frag_count++;
	} else {
		unsigned char f[HT_PAGE_SIZE];
		seek(tofs);
		int z=fread(&f, 1, HT_PAGE_SIZE, file);
		if (c!=f[ofs-tofs]) {
			d=new ht_data_page();
			for (int i=0; i<z; i++) ((ht_data_page*)d)->seto(i, f[i]);
			((ht_data_page*)d)->set(ofs-tofs, c);
			vmod->set(tofs, d);
			frag_count++;
		}
	}
	seek(ofs+1);
	if (frag_count>15) {
		frag_count=0;
		vmod->balance();
	}
}

unsigned char htrwfile::getc()
{
	ht_data *d;
	int ofs=tell();
	int tofs=ofs & ~(HT_PAGE_SIZE-1);
	if ((d=vmod->get(tofs))) {
		return ((ht_data_page*)d)->get(ofs-tofs);
	} else {
		this->seek(ofs);
		return (unsigned char)fgetc((FILE*)file);
	}
	seek(ofs+1);
}

int htrwfile::read(void *b, int size)
{
	unsigned char *buf=(unsigned char*)b;
	ht_data *d;
	int ofs=tell();
	int tofs=ofs & ~(HT_PAGE_SIZE-1);
	int z=size;
	do {
		int s=MIN(tofs+HT_PAGE_SIZE-ofs, z);
		if ((d=vmod->get(tofs))) {
			((ht_data_page*)d)->getbuf(buf, ofs-tofs, s);
		} else {
			seek(ofs);
			fread(buf, 1, s, file);
		}
		ofs+=s;
		z-=s;
		buf+=s;
		tofs+=HT_PAGE_SIZE;
	} while (z>0);
	return size;
}

int htrwfile::write(void *b, int size)
{
	unsigned char *buf=(unsigned char*)b;
	for (int i=0; i<size; i++) {
		this->putc(*buf++);
	}
	return size;
}

int htrwfile::bytedirty(int ofs)
{
	ht_data *d;
	int tofs=ofs & ~(HT_PAGE_SIZE-1);
	if ((d=vmod->get(tofs))) {
		return ((ht_data_page*)d)->dirty(ofs-tofs);
	} else return 0;
}

/*
 *	CLASS htmemfile
 */

#define HTMEMFILE_INITIAL_SIZE 1024
#define HTMEMFILE_GROW_SIZE 1024

htmemfile::htmemfile()
{
	ofs=0;
	bufsize=HTMEMFILE_INITIAL_SIZE;
	buf=(char*)malloc(bufsize);
	memset(buf, 0, bufsize);
	pos=0;
	isdirty=0;
}

htmemfile::htmemfile(int ofs, int size)
{
	this->ofs=ofs;
	buf=(char*)malloc(size);
	bufsize=size;
	memset(buf, 0, size);
	pos=0;
	isdirty=0;
}

htmemfile::~htmemfile()
{
	delete buf;
}

int htmemfile::read(void *buf, int size)
{
	if (pos+size>bufsize) {
		memset(buf, 0, size);
		if (bufsize-pos+1>0) memmove(buf, this->buf+pos, bufsize-pos+1);
	} else {
		memmove(buf, (char*)this->buf+pos, size);
	}
	pos+=size;
	return size;
}

int htmemfile::write(void *buf, int size)
{
	isdirty=1;
	while (pos+size>bufsize) extendbuf();
	memmove(this->buf, buf, size);
	return size;
}

int htmemfile::seek(int pos)
{
	if (pos<ofs)
		return 0;
	if (pos-ofs>bufsize)
		return 0;
	this->pos=pos-ofs;
	return 1;
}

int htmemfile::tell()
{
	return pos+ofs;
}

int htmemfile::size()
{
	return bufsize;
}

char *htmemfile::getname()
{
	return "memfile";
}

void htmemfile::extendbuf()
{
	bufsize+=HTMEMFILE_GROW_SIZE;
	char *t=(char*)malloc(bufsize);
	memset(t, 0, bufsize);
	memmove(t, buf, bufsize-HTMEMFILE_GROW_SIZE);
	delete buf;
	buf=t;
}

int htmemfile::dirty()
{
	int d=isdirty;
	isdirty=0;
	return d;
}

/*
 *	string file functions
 */

char *getstrz(htfile *file)
{
/* get string size */
	char buf[64];
	int s, z=0, o=file->tell();
	int found=0;
	while (!found) {
		if (!(s=file->read(buf, 64))) break;
		for (int i=0; i<s; i++) {
			z++;
			if (!buf[i]) {
				found=1;
				break;
			}
		}
	}
/* read string */
	file->seek(o);
	char *str=(char *)malloc(z);
	file->read(str, z);
	return str;
}

void putstrz(htfile *file, char *str)
{
	file->write(str, strlen(str)+1);
}

char *getstrl(htfile *file)
{
	unsigned char l;
	file->read(&l, 1);
	char *str=(char*)malloc(l+1);
	file->read(str, l);
	*(str+l)=0;
	return str;
}

void putstrl(htfile *file, char *str)
{
	unsigned char l=strlen(str);
	file->write(&l, 1);
	file->write(str, l);
}

