/*
 *   htcurses.cc - LINUX implementation
 */

#include "config.h"
#include <unistd.h>

#include "global.h"
#include "htcurses.h"
#include "htdebug.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *	CLASS screendrawbuf
 */
 
drawbufch *buf;
WINDOW *win;
int use_colors, use_high_colors;
int is_xterm;
int cursorx, cursory; 

screendrawbuf::screendrawbuf()
{
	bounds b;
	buf=0;
	int colors[8]={ COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN, COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE };

	cursorx=0;
	cursory=0;
	win=initscr();
	use_colors=0;
	use_high_colors=0;
	if (has_colors()) {
	    char *term=getenv("TERM");
	    int bold_support=0;
	    start_color();
	    use_colors=1;
/* FIXME: Does this work ???: test if the A_BOLD attr can be set */
	    attrset(A_BOLD);
	    bold_support=((attr_get() & A_BOLD)!=0);
	    attrset(A_NORMAL);
	    
	    is_xterm=(strcmp(term, "linux") && strcmp(term, "console"));
	    if ( (!is_xterm) && (!bold_support)) {
		WARN("terminal is of type '%s' (non-x-terminal) but bold_test fails !", term);
	    }
	    if (!is_xterm || bold_support) {
		    use_high_colors=1;
	    } else {
		    WARN("terminal only supports up to 8 foreground colors !");
	    }
	    for (int fg=0; fg<8; fg++) {
		    for (int bg=0; bg<8; bg++) {
			init_pair(fg+(bg<<3)+1, colors[fg], colors[bg]);
		    }
	    }
	} else {
	    WARN("terminal lacks color support !");
	}
	timeout(0);
	meta(win, 1);
	keypad(win, 1);
	nodelay(win, 1);
	noecho();
	ESCDELAY=0;
	
	b.x=0;
	b.y=0;
	b.w=getmaxx(win);
	b.h=getmaxy(win);
	setbounds(&b);

	show();
}

screendrawbuf::~screendrawbuf()
{
	endwin();
	if (buf) delete buf;
}

int screendrawbuf::print(int x, int y, int c, char *text)
{
	return lprint(x, y, c, 0x7fffffff, text);
}

void screendrawbuf::printchar(int x, int y, int c, int ch)
{
	drawbufch *b=buf+x+y*size.w;
	b->ch=ch;
	b->c=c;
}

int screendrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int n=0;
	drawbufch *ch=buf+x+y*size.w;
	while ((*text) && (n<l)) {
		ch->ch=*text;
		ch->c=c;
		ch++;	
		text++;	
		n++;
	}
	return n;
}

void screendrawbuf::fill(int x, int y, int w, int h, int c, int chr)
{
	for (int i=0; i<h; i++) {
		drawbufch *ch=buf+x+(y+i)*size.w;
		for (int j=0; j<w; j++) {
			ch->ch=chr;
			ch->c=c;
			ch++;
		}
	}
}

void screendrawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) delete buf;
	buf=(drawbufch *)malloc(sizeof *buf * size.w * size.h);
	fill(size.x, size.y, size.w, size.h, 7, ' ');
}

void screendrawbuf::drawbuffer(drawbuf *b, int x, int y)
{
	for (int iy=0; iy<b->size.h; iy++) {
		drawbufch *src=b->buf+iy*b->size.w;
		drawbufch *dest=buf+x+(y+iy)*size.w;
		memmove(dest, src, b->size.w * sizeof *dest);
	}
}

void screendrawbuf::show()
{
	drawbufch *ch=buf;
	int c=-1;
	for (int iy=0; iy<size.h; iy++) {
			move(iy+size.y, size.x);
			for (int ix=0; ix<size.w; ix++) {
				if ((use_colors) && (ch->c!=c)) {
					c=ch->c;
					if (use_high_colors) {
						if (is_xterm && (c==8)) {
					    /* some terminals can't display grey (bold black) correctly... */
						    c=7;
						} else {
						    if ((c&15)>7) attrset(A_BOLD); else attrset(A_NORMAL);
						}						    
					}					    
					bkgdset(COLOR_PAIR( (c&7) + (((c>>4)&7)<<3) + 1 ));
				}
				if ((((unsigned char)ch->ch>=0x20) && ((unsigned char)ch->ch<=0x7e)) || ((unsigned int)ch->ch>0xff)) addch(ch->ch); else addch(32);
				ch++;
			}
	}
	setsyx(cursory, cursorx);
	refresh();
}

void screendrawbuf::getcursor(int *x, int *y)
{
	*x=cursorx;
	*y=cursory;
}

void screendrawbuf::hidecursor()
{
	setcursor(0, 0);
}

void screendrawbuf::setcursor(int x, int y)
{
	cursorx=x;
	cursory=y;
}

void screendrawbuf::showcursor()
{
}

/*
 *	CLASS drawbuf
 */

drawbuf::drawbuf(bounds *b)
{
	buf=0;
	setbounds(b);
}

drawbuf::~drawbuf()
{
	if (buf) {
		delete buf;
	}
}

int drawbuf::print(int x, int y, int c, char *text)
{
	return lprint(x, y, c, 0x7fffffff, text);
}

void drawbuf::printchar(int x, int y, int c, int ch)
{
	drawbufch *b=buf+(x-size.x)+(y-size.y)*size.w;
	b->ch=ch;
	b->c=c;
}

int drawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int ox=0;
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	if (y<size.h) {
		while ((ox<size.w) && (*text) && (ox<l)) {
			b->ch=(byte)*text;
			b->c=c;
			text++;
			ox++;
			b++;
		}
	}
	return ox;
}

void drawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
	x-=size.x;
	y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	for (int iy=y; iy<y+h; iy++) {
		for (int ix=x; ix<x+w; ix++) {
			b->ch=ch;
			b->c=c;
			b++;
		}
	}
}

void drawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) delete buf;
	buf=(drawbufch*)malloc(sizeof *buf * size.w * size.h);
	fill(size.x, size.y, size.w, size.h, 7, ' ');
}

/*****************************************************************************
 ****************************************************************************/

genericdrawbuf::genericdrawbuf()
{
}

genericdrawbuf::~genericdrawbuf()
{
}

void genericdrawbuf::fill(int x, int y, int w, int h, int c, int ch)
{
}

int genericdrawbuf::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return print(x, y, c, buf);
}

int genericdrawbuf::print(int x, int y, int c, char *text)
{
	return 0;
}

void genericdrawbuf::printchar(int x, int y, int c, int ch)
{
}

int genericdrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	return 0;
}

void genericdrawbuf::setbounds(bounds *b)
{
	size=*b;
}


